@@ -6,7 +6,7 @@ use Mojo::JSON;
 use Mojo::Util 'deprecated';
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 sub EXCEPTION { +{errors => [{message => 'Internal server error.', path => '/'}], status => 500} }
 sub NOT_FOUND { +{errors => [{message => 'Not found.',             path => '/'}], status => 404} }
@@ -21,11 +21,10 @@ sub register {
   my $api_spec = $self->_load_spec($app, $config);
 
   unless ($app->defaults->{'openapi.base_paths'}) {
-    $app->helper('openapi.invalid_input' => \&_invalid_input);
-    $app->helper('openapi.validate'      => \&_validate);
-    $app->helper('openapi.valid_input'   => \&_valid_input);
-    $app->helper('openapi.spec'          => \&_helper_spec);
-    $app->helper('reply.openapi'         => \&_reply);
+    $app->helper('openapi.validate'    => \&_validate);
+    $app->helper('openapi.valid_input' => sub { _validate($_[0]) ? undef : $_[0] });
+    $app->helper('openapi.spec'        => \&_helper_spec);
+    $app->helper('reply.openapi'       => \&_reply);
     $app->hook(before_render => \&_before_render);
     $app->renderer->add_handler(openapi => \&_render);
     push @{$app->renderer->classes}, __PACKAGE__;
@@ -61,6 +60,7 @@ sub _add_routes {
 
   for my $path (sort { length $a <=> length $b } keys %$paths) {
     next if $path =~ $X_RE;
+    my $has_options;
 
     for my $http_method (keys %{$paths->{$path}}) {
       next if $http_method =~ $X_RE;
@@ -69,6 +69,8 @@ sub _add_routes {
       my $to      = $op_spec->{'x-mojo-to'};
       my $endpoint;
 
+      $has_options = 1 if lc $http_method eq 'options';
+
       die qq([OpenAPI] operationId "$op_spec->{operationId}" is not unique)
         if $op_spec->{operationId} and $uniq{o}{$op_spec->{operationId}}++;
       die qq([OpenAPI] Route name "$name" is not unique.) if $name and $uniq{r}{$name}++;
@@ -85,6 +87,10 @@ sub _add_routes {
       $endpoint->to({'openapi.op_spec' => $op_spec});
       warn "[OpenAPI] Add route $http_method @{[$endpoint->render]}\n" if DEBUG;
     }
+
+    unless ($has_options) {
+      $route->options($path => sub { _render_route_spec($_[0], $path) });
+    }
   }
 }
 
@@ -109,11 +115,6 @@ sub _helper_spec {
   return $c->stash('openapi.api_spec')->get($path);
 }
 
-sub _invalid_input {
-  deprecated 'Use $c->openapi->validate or $c->openapi->valid_input instead';
-  goto &_validate;
-}
-
 sub _load_spec {
   my ($self, $app, $config) = @_;
   my $openapi = JSON::Validator->new->schema(JSON::Validator::OpenAPI::SPECIFICATION_URL());
@@ -185,6 +186,15 @@ sub _render {
   }
 }
 
+sub _render_route_spec {
+  my ($c, $path) = @_;
+  my $spec   = $c->stash('openapi.api_spec')->data->{paths}{$path};
+  my $method = $c->param('method');
+  $spec = $spec->{$method} if $method;
+  return $c->render(json => $spec) if $spec;
+  return $c->render(json => {}, status => 404);
+}
+
 sub _reply_spec {
   my $c      = shift;
   my $spec   = $c->stash('openapi.api_spec')->data;
@@ -224,7 +234,9 @@ sub _validate {
   my ($c, $args) = @_;
   my $self    = $c->stash('openapi.object');
   my $op_spec = $c->openapi->spec;
-  my @errors  = $self->_validator->validate_request($c, $op_spec, $c->validation->output);
+
+  # Write validated data to $c->validation->output
+  my @errors = $self->_validator->validate_request($c, $op_spec, $c->validation->output);
 
   if (@errors) {
     $self->_log($c, '<<<', \@errors);
@@ -235,8 +247,6 @@ sub _validate {
   return @errors;
 }
 
-sub _valid_input { _validate($_[0], {}) ? undef : $_[0]; }
-
 1;
 
 =encoding utf8
@@ -357,12 +367,6 @@ Returns the L<Mojolicious::Controller> object if the input is valid or
 automatically render an error document if not and return false. See
 L</SYNOPSIS> for example usage.
 
-=head2 reply.openapi
-
-Will be DEPRECATED in favor of:
-
-  $c->render(openapi => $output);
-
 =head1 METHODS
 
 =head2 register
